import socket
import ssl

from prestige_irc import connection
from prestige_irc.commands import Commands
from prestige_irc.connection import MessageListener
from prestige_irc.message import IRCMessage


class IRCConnection(connection.Connection):

    """Creates a connection to an IRC network."""

    def __init__(self, nick):
        """
        Readies the connection to the irc network, and sets up the initial nick name to use.

        Parameters
        ----------
        nick: str
            The irc nick name to use.
        """
        super().__init__()
        # Set the local user nickname.
        self.__nick = nick
        # Listener which automatically handles ping responses.
        self.add_listener(MessageListener(message_filter=lambda conn, msg: msg.command == Commands.PING,
                                          receive=lambda conn, msg: conn.cmd_pong(msg.target)))

    def _process_data(self, data):
        """
        Processes the bytes that are received from the server, and converts them into an IRCMessage.

        Parameters
        ----------
        data: bytes
            The bytes to convert into an `IRCMessage`.
        """
        return IRCMessage(data.decode('utf-8'))

    def connect(self, ip_address, port=6697, timeout=None, enable_ssl=True):
        """
        Attempts to connect to the specified IP address and port.

        Parameters
        ----------
        ip_address: str
            The IP address to connect to.
        port: int
            The port number to bind to.
            Default value is 6697, which is the default port for TLS/SSL connections
            (See https://datatracker.ietf.org/doc/rfc7194/ for details).
        timeout: int|None (optional)
            The number of seconds to wait to stop attempting to connect if a connection has not yet been made.
            Default value is None.
        enable_ssl: bool (optional)
            If the connection should be made with SSL.
            Default value is True.

        Returns
        -------
        bool:
            If the connection was successfully established.
        """
        if enable_ssl:
            return self.__connect_ssl(ip_address=ip_address, port=port)
        else:
            return self.__connect(ip_address=ip_address, port=port)

    def __connect(self, ip_address, port, timeout=None):
        """
        Attempts to connect to the specified IP address and port.

        Parameters
        ----------
        ip_address: str
            The IP address to connect to.
        port: int
            The port number to bind to.
        timeout: int|None (optional)
            The number of seconds to wait to stop attempting to connect if a connection has not yet been made.
            Default value is None.

        Returns
        -------
        bool:
            If the connection was successfully established.
        """
        connection_successful = super().connect(ip_address=ip_address, port=port, timeout=timeout)
        if connection_successful:
            self.__on_connect()
        return connection_successful

    def __connect_ssl(self, ip_address, port, timeout=None):
        """
        Attempts to connect to the specified IP address and port.

        Parameters
        ----------
        ip_address: str
            The IP address to connect to.
        port: int
            The port number to bind to.
        timeout: int|None (optional)
            The number of seconds to wait to stop attempting to connect if a connection has not yet been made.
            Default value is None.

        Returns
        -------
        bool:
            If the connection was successfully established.
        """
        sock = socket.socket()
        sock.settimeout(timeout)
        ssl_socket = ssl.create_default_context().wrap_socket(sock=sock, server_hostname=ip_address)
        return self.connect_socket(sock=ssl_socket, ip_address=ip_address, port=port)

    def connect_socket(self, sock, ip_address, port):
        """Connect to a server.

        Parameters
        ----------
        sock: SSLSocket
            The socket to connect with.
        ip_address: str
            The IP address of the server.
        port: int
            The port number to bind to.

        Returns
        -------
        bool:
            If the connection was successfully established.

        Throws
        ------
        socket.error:
            If a connection could not be made.
        """
        connection_successful = super().connect_socket(sock=sock, ip_address=ip_address, port=port)
        if connection_successful:
            self.__on_connect()
        return connection_successful

    def __on_connect(self):
        """Runs IRC commands needed after the connection has been established."""
        self.cmd_nick(nick=self.__nick)
        self.cmd_user(real_name=self.__nick)

    @property
    def nick(self):
        """
        Gets the user's nick.

        The internal property is initialized in the constructor, changed with the use of `cmd_nick`,
        and should not be modified by other means.

        Returns
        -------
        str:
            The user's nick.
        """
        return self.__nick

    # --------------------------- #
    # IRC Commands Implementation #
    # --------------------------- #

    def send_command(self, command, prefix='', params=''):
        """
        Sends commands to the server. All functions prefixed with 'cmd' pass through this method.

        The message sent will be:
            prefix + command + " " + params

        Parameters
        ----------
        command: str
            The irc command, which can be found in commands.Commands.
        prefix: str (optional)
            A prefix to the command.
            Default value is an empty string.
        params: str (optional)
            The parameters of the command, in one string.
            Default value is an empty string.

        Returns
        -------
        bool:
            If the message was sent successfully.
            If False is returned, this typically means the connection has been terminated.
        """
        if self.is_connection_alive:
            self.send(f'{prefix}{command}' + f' {params}' if params else '')
            return True
        return False

    def cmd_admin(self, target=''):
        """
        Instructs the server to return information about the administrators of the server specified by `target`,
        where `target` is either a server or a user. If `target` is omitted, the server should return information
        about the administrators of the current server.

        Parameters
        ----------
        target: str (optional)
            A server or a user.
            Default value is an empty string.
        """
        self.send_command(command=Commands.ADMIN, params=target)

    def cmd_away(self, message=''):
        """
        Provides the server with a `message` to automatically send in reply to a PRIVMSG directed at the user,
        but not to a channel they are on. If `message` is omitted, the away status is removed.

        Parameters
        ----------
        message: str (optional)
            The away message to send to the server.
            Default value is an empty string.
        """
        self.send_command(command=Commands.AWAY, params=message)

    def cmd_cnotice(self, nickname, channel, message):
        """
        Sends a channel NOTICE message to `nickname` on `channel` that bypasses flood protection limits.
        The target nickname must be in the same channel as the client issuing the command,
        and the client must be a channel operator.

        Normally an IRC server will limit the number of different targets a client can send messages to within
        a certain time frame to prevent spammers or bots from mass-messaging users on the network,
        however this command can be used by channel operators to bypass that limit in their channel.
        For example, it is often used by help operators that may be communicating with a large number of users
        in a help channel at one time.

        This command is not formally defined in an RFC, but is in use by some IRC networks.
        Support is indicated in a RPL_ISUPPORT reply (numeric 005) with the CNOTICE keyword.

        Parameters
        ----------
        nickname: str
            The nick of the user to send the notice to.
        channel: str
            The channel the user is on to send the notice through.
        message: str
            The notice message to send to the user.
        """
        self.send_command(command=Commands.CNOTICE,
                          params=f'{nickname} {channel} :{message}')

    def cmd_cprivmsg(self, nickname, channel, message):
        """
        Sends a private message to `nickname` on `channel` that bypasses flood protection limits.
        The target nickname must be in the same channel as the client issuing the command,
        and the client must be a channel operator.

        Normally an IRC server will limit the number of different targets a client can send messages to within
        a certain time frame to prevent spammers or bots from mass-messaging users on the network,
        however this command can be used by channel operators to bypass that limit in their channel.
        For example, it is often used by help operators that may be communicating with a large number of users
        in a help channel at one time.

        This command is not formally defined in an RFC, but is in use by some IRC networks.
        Support is indicated in a RPL_ISUPPORT reply (numeric 005) with the CPRIVMSG keyword.

        Parameters
        ----------
        nickname: str
            The nick of the user to send the message to.
        channel: str
            The channel the user is on to send the message through.
        message: str
            The message to send the user.
        """
        self.send_command(command=Commands.CPRIVMSG,
                          params=f'{nickname} {channel} :{message}')

    def cmd_connect(self, target_server, port, remote_server=None):
        """
        Instructs the server `remote server` (or the current server, if `remote server` is omitted)
        to connect to `target server` on port `port`.
        This command should only be available to IRC Operators.

        Parameters
        ----------
        target_server: str
            The name of the server to connect the remove_server to.
        port: int
            The port number to connect to.
        remote_server: str (optional)
            The remote server to connect the target sever to.
            If omitted, this parameter will use the current server.
        """
        self.send_command(command=Commands.CONNECT,
                          params=f'{target_server} {str(port)}' + f' {remote_server}' if remote_server else '')

    def cmd_die(self):
        """
        This command may only be issued by IRC server operators.
        Instructs the server to shut down.

        Parameters
        ----------
        """
        self.send_command(command=Commands.DIE)

    def cmd_encap(self, destination, subcommand, parameters):
        """
        This command is for use by servers to encapsulate commands so that they will propagate across
        hub servers not yet updated to support them, and indicates the subcommand and its parameters should be passed
        unaltered to the destination, where it will be unencapsulated and parsed.
        This facilitates implementation of new features without a need to restart all servers before they are
        usable across the network.

        Parameters
        ----------
        destination: str
            The hub server destination.
        subcommand: str
            The command to send.
        parameters: str
            The parameters of the command being sent.
        """
        self.send_command(command=Commands.ENCAP,
                          params=f'{destination} {subcommand} {parameters}')

    def cmd_error(self, error_message):
        """
        This command is for use by servers to report errors to other servers.
        It is also used before terminating client connections.

        Parameters
        ----------
        error_message: str
            The error message to send.
        """
        self.send_command(command=Commands.ERROR, params=error_message)

    def cmd_help(self):
        """
        Requests the server help file.
        This command is not formally defined in an RFC, but is in use by most major IRC daemons.
        """
        self.send_command(command=Commands.HELP)

    def cmd_info(self, target=''):
        """
        Returns information about the `target` server, or the current server if `target` is omitted.
        Information returned includes the server's version, when it was compiled, the patch level, when it was started,
        and any other information which may be considered to be relevant.

        Parameters
        ----------
        target: str (optional)
            The target server to request information from.
            Default value is an empty string.
        """
        self.send_command(command=Commands.INFO, params=target)

    def cmd_invite(self, nickname, channel):
        """
        Invites `nickname` to the channel `channel`. `channel` does not have to exist, but if it does,
        only members of the channel are allowed to invite other clients.
        If the channel mode i is set, only channel operators may invite other clients.

        Parameters
        ----------
        nickname: str
            The nickname of the user to invite.
        channel: str
            The channel to invite the user to.
        """
        self.send_command(command=Commands.INVITE, params=f'{nickname} {channel}')

    def cmd_ison(self, nicknames):
        """
        Queries the server to see if the clients in the list `nicknames` are currently on the network.
        The server returns only the nicknames that are on the network in a list.
        If none of the clients are on the network, the server returns an empty list.

        Parameters
        ----------
        nicknames: list
            A list of nicknames.
        """
        self.send_command(command=Commands.ISON,
                          params=' '.join(nick for nick in nicknames))

    def cmd_join(self, channels):
        """
        Joins the specified channels.

        Parameters
        ----------
        channels: collections.iterable
            A list of channels, prefixed with `#`
            This method automatically adds a `#` to the channel name if it is absent.
        """
        self.send_command(command=Commands.JOIN,
                          params=','.join(f'#{channel}' if channel[0] != '#' else channel for channel in channels))

    def cmd_kick(self, channel, nickname, message=''):
        """
        Forcibly removes `nickname` from `channel`.
        This command may only be issued by channel operators.

        Parameters
        ----------
        channel: str
            The channel to kick the user from.
        nickname: str
            The nickname of the client to kick from the channel.
        message: str (optional)
            The message to send to the user about being kicked from the channel.
            Default value is an empty string.
        """
        self.send_command(command=Commands.KICK,
                          params=f'{channel} {nickname}' + f' {message}' if message else '')

    def cmd_kill(self, nickname, message):
        """
        Forcibly removes `nickname` from the network. This command may only be issued by IRC operators.

        Parameters
        ----------
        nickname: str
            The nickname of the user to remove from the network.
        message: str
            The reason for the kill command, sent to the user.
        """
        self.send_command(command=Commands.KILL, params=f'{nickname} {message}')

    def cmd_knock(self, channel, message=''):
        """
        Sends a NOTICE to an invitation-only `channel` with an optional `message`, requesting an invite.
        This command is not formally defined by an RFC, but is supported by most major IRC daemons.
        Support is indicated in a RPL_ISUPPORT reply (numeric 005) with the KNOCK keyword.

        Parameters
        ----------
        channel: str
            The channel to send the request to.
        message: str (optional)
            The message to send with the request.
        """
        self.send_command(command=Commands.KNOCK, params=channel + f' {message}' if message else '')

    def cmd_links(self, remote_server='', server_mask=''):
        """
        Lists all server links matching `server mask`, if given, on `remote server`, or the current server if omitted.

        Parameters
        ----------
        remote_server: str (optional)
            The server to check. If omitted, the current server is used.
            Default value is an empty string.
        server_mask: str (optional)
            The mask used to check for server links. Lists all links if omitted.
            Default value is an empty string.
        """
        self.send_command(command=Commands.LINKS, params=f'{remote_server} ' if remote_server else server_mask)

    def cmd_list(self, channels=None, server=''):
        """
        Lists all channels on the server.
        If the list `channels` is given, it will return the channel topics.
        If `server` is given, the command will be forwarded to `server` for evaluation.

        Parameters
        ----------
        channels: list (optional)
            The channels to get the topics from.
            Default value is None.
        server: str (optional)
            The server to send the topic request to.
            Default value is an empty string.
        """
        self.send_command(command=Commands.LIST,
                          params=','.join(channels) + ' ' if channels else server)

    def cmd_lusers(self, mask='', target=''):
        """
        The LUSERS command is used to get statistics about the size of the IRC network.
        If no parameter is given, the reply will be about the whole net.
        If a `mask` is specified, then the reply will only concern the part of the network
        formed by the servers matching the mask.
        Finally, if the `target` parameter is specified, the request is forwarded
        to that server which will generate the reply.

        Parameters
        ----------
        mask: str (optional)
            The mask used to specify a certain part of the network.
            The default value is an empty string.
        target: str (optional)
            The target server to send the request to.
            The default value is an empty string.
        """
        self.send_command(command=Commands.LUSERS,
                          params=f'{mask} ' if mask else target)

    def cmd_mode_channel(self, channel, flags, params=''):
        """
        The MODE command is provided so that channel operators may change the characteristics of `their' channel.
        It is also required that servers be able to change channel modes so that channel operators may be created.

        The various modes available for channels are as follows:

            o - give/take channel operator privileges;
            p - private channel flag;
            s - secret channel flag;
            i - invite-only channel flag;
            t - topic settable by channel operator only flag;
            n - no messages to channel from clients on the outside;
            m - moderated channel;
            l - set the user limit to channel;

        When using the 'o' and 'b' options, a restriction on a total of three per mode command has been imposed.

        Parameters
        ----------
        channel: str
            The channel of which the mode is being set.
        flags: str
            The mode flags to set.
        params: str (optional)
            Optional parameters for the command; see RFC for specification.
            Default value is an empty string.
        """
        self.send_command(command=Commands.MODE,
                          params=f'{channel} {flags}' + f' {params}' if params else '')

    def cmd_mode_nickname(self, nickname, flags, params=''):
        """
        The user MODEs are typically changes which affect either how the
        client is seen by others or what 'extra' messages the client is sent.
        A user MODE command may only be accepted if both the sender of the
        message and the nickname given as a parameter are both the same.

        The available modes are as follows:

            i - marks a users as invisible;
            s - marks a user for receipt of server notices;
            w - user receives wallops;
            o - operator flag.

        Parameter
        ---------
        nickname: str
            The nickname of the user.
        flags: str
            The mode flags to set.
        params: str (optional)
            Optional parameters for the command; see RFC for specification.
            Default value is an empty string.
        """
        self.send_command(command=Commands.MODE,
                          params=f'{nickname} {flags}' + f' {params}' if params else '')

    def cmd_motd(self, server=''):
        """
        Returns the message of the day on `server` or the current server if it is omitted.

        Parameters
        ----------
        server: str (optional)
            The server to retrieve the message from, or the current server if omitted.
        """
        self.send_command(command=Commands.MOTD, params=server)

    def cmd_names(self, channels=None, server=''):
        """
        Returns a list of who is on the comma-separated list of `channels`, by channel name.
        If `channels` is omitted, all users are shown, grouped by channel name with all users who are not on a channel
        being shown as part of channel "*".
        If `server` is specified, the command is sent to `server` for evaluation.

        Parameters
        ----------
        channels: list (optional)
            The channels from which to return the list of user nicks.
            If omitted, all users are returned.
        server: str (optional)
            If `server` is specified, the command is sent to `server` for evaluation.
        """
        self.send_command(command=Commands.NAMES,
                          params=','.join(channel for channel in channels) +
                                 (f' {server}' if server else '') if channels is not None else '')

    def cmd_nick(self, nick):
        """
        Attempts to set the user's nick on the irc server.

        Parameters
        ----------
        nick: str
            The user's nick name.
        """
        self.__nick = nick
        self.send_command(command=Commands.NICK, params=nick)

    def cmd_privmsg(self, target, message):
        """
        Send a message to a channel or a user.

        Parameters
        ----------
        target: str
            The user or channel to send a message to. If a channel, it must be prefixed with '#'.
        message: str
            The message to send.
        """
        self.send_command(command=Commands.PRIVMSG, params=f'{target} :{message}')

    def cmd_user(self, real_name, invisible=False):
        """
        Sends the user data to the server. This is done automatically upon calling IRCConnection#connect,
        using IRCConnection#nick as the default for real_name.

        Parameters
        ----------
        real_name: str
            The user's "real name" which is visible to other members on the irc network.
        invisible: bool (optional)
            If the user wishes to remain invisible to other members on the network, aside from other members in the same
            channel(s).
            Default value is False.
        """
        self.send_command(command=Commands.USER, params=f'{self.__nick} {str(8 if invisible else 0)} * :{real_name}')

    def cmd_part(self, channels, reason=''):
        """
        Leaves the specified channels.

        Parameters
        ----------
        channels: collections.iterable
            A list of channels, prefixed with '#'
            This method automatically adds a '#' to the channel name if it is absent.
        reason: str (optional
            The reason for leaving the channel(s).
            Default value is an empty string.
        """
        self.send_command(command=Commands.PART,
                          params=','.join((f'#{channel}' if channel[0] != '#' else channel)
                                          for channel in channels) + f' :{reason}')

    def cmd_pong(self, message):
        """
        Sends a pong message to the server

        Parameters
        ----------
        message: str
            The argument after "PING" sent from the server.
        """
        self.send_command(command=Commands.PONG, params=f':{message}')

    def cmd_quit(self, reason=''):
        """
        Terminates the connection with the server, and sends an optional reason for quitting.

        Parameters
        ----------
        reason: str (optional)
            The reason for terminating the connection.
            Default is an empty string.
        """
        self.send_command(command=Commands.QUIT, params=reason)



class Commands:

    """A list of IRC commands, which are used in `IRCConnection`.

    This list of commands is based on:
        https://en.wikipedia.org/wiki/List_of_Internet_Relay_Chat_commands
    """

    ADMIN = 'ADMIN'
    AWAY = 'AWAY'
    CNOTICE = 'CNOTICE'
    CPRIVMSG = 'CPRIVMSG'
    CONNECT = 'CONNECT'
    DIE = 'DIE'
    ENCAP = 'ENCAP'
    ERROR = 'ERROR'
    HELP = 'HELP'
    INFO = 'INFO'
    INVITE = 'INVITE'
    ISON = 'ISON'
    JOIN = 'JOIN'
    KICK = 'KICK'
    KILL = 'KILL'
    KNOCK = 'KNOCK'
    LINKS = 'LINKS'
    LIST = 'LIST'
    LUSERS = 'LUSERS'
    MODE = 'MODE'
    MOTD = 'MOTD'
    NAMES = 'NAMES'
    NAMESX = 'NAMESX'
    NICK = 'NICK'
    NOTICE = 'NOTICE'
    OPER = 'OPER'
    PART = 'PART'
    PASS = 'PASS'
    PING = 'PING'
    PONG = 'PONG'
    PRIVMSG = 'PRIVMSG'
    QUIT = 'QUIT'
    REHASH = 'REHASH'
    RESTART = 'RESTART'
    RULES = 'RULES'
    SERVER = 'SERVER'
    SERVICE = 'SERVICE'
    SERVLIST = 'SERVLIST'
    SQUERY = 'SQUERY'
    SQUIT = 'SQUIT'
    SETNAME = 'SETNAME'
    SILENCE = 'SILENCE'
    STATS = 'STATS'
    SUMMON = 'SUMMON'
    TIME = 'TIME'
    TOPIC = 'TOPIC'
    TRACE = 'TRACE'
    UHNAMES = 'UHNAMES'
    USER = 'USER'
    USERHOST = 'USERHOST'
    USERIP = 'USERIP'
    USERS = 'USERS'
    VERSION = 'VERSION'
    WALLOPS = 'WALLOPS'
    WATCH = 'WATCH'
    WHO = 'WHO'
    WHOIS = 'WHOIS'
    WHOWAS = 'WHOWAS'



def parse(raw_message):
    """Breaks a message from an IRC server into components.

    Parameters
    ----------
    raw_message: str
        The raw message received from the IRC server.

    Returns
    -------
    nick: str
        The nick name of the sender.
    host: str
        The host of the IRC message (nick!user@host).
    command: str
        The IRC command.
    target: str
        The target to which the message was sent (usually #channel or nick).
    text: str
        The text sent (the last value in `args`).
    args: list
        The arguments in the IRC message.
    """
    host = ''
    if not raw_message:
        raise Exception('Cannot parse an empty message.')
    if raw_message[0] == ':':
        host, raw_message = raw_message[1:].split(' ', 1)
    if raw_message.find(' :') != -1:
        raw_message, trailing = raw_message.split(' :', 1)
        args = raw_message.split()
        args.append(trailing)
    else:
        args = raw_message.split()

    command = args.pop(0)
    nick = host.split('!', 2)[0] if '!' in host else ''
    target = args[0] if len(args) > 0 else ''
    text = args[-1] if len(args) > 0 else ''
    return nick, host, command, target, text, args


class IRCMessage(object):

    def __init__(self, raw_message):
        """
        Parses the raw string received from the server and
        separates each piece into a field using IRC.parse(raw_message).

        Parameters
        ----------
        raw_message: str
            The raw string received from the server.
        """
        # Parse the raw message.
        self.raw = raw_message
        self.nick, self.host, self.command, self.target, self.text, self.args = parse(raw_message)

    def __str__(self):
        return 'Raw: ' + self.raw + \
            '\r\nNick: ' + str(self.nick) + \
            '\r\nHost: ' + str(self.host) + \
            '\r\nCommand: ' + str(self.command) + \
            '\r\nTarget: ' + str(self.target) + \
            '\r\nText: ' + str(self.text) + \
            '\r\nArgs: ' + str(self.args)



from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="prestige_irc",
    version="0.0.6",
    author="Avahe Kellenberger",
    author_email="avahe@protonmail.ch",
    description="A simple API for IRC networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/avahe-kellenberger/prestige_irc",
    packages=["prestige_irc"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent"
    ]
)



import socket
import threading


class Connection(object):

    """
    A class for basic generic connectivity.
    """

    def __init__(self):
        """
        Readies a connection to a server at a specific port, and keeps the connection alive.
        """
        self.__socket = None
        self.__is_connection_alive = False
        self.__listen_thread = None
        self.__listeners = set()

    def connect(self, ip_address, port, timeout=None):
        """Connect to a server.

        Parameters
        ----------
        ip_address: str
            The IP address of the server.
        port: int
            The port number to bind to.
        timeout: int|None
            The number of seconds to wait to stop attempting to connect if a connection has not yet been made.
            Default value is None.

        Returns
        -------
        bool:
            If the connection was successfully established.

        Throws
        ------
        socket.error:
            If a connection could not be made.
        """
        sock = socket.socket()
        sock.settimeout(timeout)
        return self.connect_socket(sock=sock, ip_address=ip_address, port=port)

    def connect_socket(self, sock, ip_address, port):
        """Connect to a server using the given socket.

        Parameters
        ----------
        sock: socket.py
            The unconnected socket with which to establish the connection.
        ip_address: str
            The IP address of the server.
        port: int
            The port number to bind to.

        Returns
        -------
        bool:
            If the connection was successfully established.

        Throws
        ------
        socket.error:
            If a connection could not be made.
        """
        if not self.__is_connection_alive:
            try:
                self.__socket = sock
                self.__socket.connect((ip_address, port))
                self.__is_connection_alive = True
                self.__listen_thread = threading.Thread(target=self.__listen)
                self.__listen_thread.start()
            except socket.error as err:
                print(f'Caught exception socket.error:\n{str(err)}')
                self.__is_connection_alive = False
            return self.__is_connection_alive

        return False

    def disconnect(self):
        """Disconnects from the server.

        Returns
        -------
        bool:
           If the connection was successfully terminated.
        """
        if self.__is_connection_alive:
            self.__socket.close()
            self.__is_connection_alive = False
            return True
        return False

    @property
    def is_connection_alive(self):
        """Checks if connection is still live.

        Returns
        -------
        bool:
            If the connection is currently connected.
        """
        return self.__is_connection_alive

    def send_data(self, data):
        """Sends bytes across the connection.

        Parameters
        ----------
        data: bytes
            The bytes to send.
        """
        self.__socket.send(data)

    def send(self, message, crlf_ending=True):
        """Helper function; sends a string across the connection as bytes.

        Parameters
        ----------
        message: str
            The message to send across the connection.
        crlf_ending: bool
            If the method should ensure the message ends with CR-LF.
            Default value is True.
        """
        self.send_data(bytes(f'{message}\r\n' if crlf_ending and not message.endswith('\r\n') else message, 'utf-8'))

    def add_listener(self, listener):
        """Adds a listener to the connection.

        Parameters
        ----------
        listener:
            A function which accepts an object which is created from Connection._process_data(bytes),
            which is called each time data is received from the server.
        """
        self.__listeners.add(listener)

    def remove_listener(self, listener):
        """Removes a listener from the connection.

        Parameters
        ----------
        listener:
            A function which accepts an object which is created from Connection._process_data(bytes),
            which is called each time data is received from the server.
        """
        self.__listeners.remove(listener)

    def _process_data(self, data):
        """Processes the bytes received by the server.

        Parameters
        ----------
        data: bytes
            The bytes received from the server.

        Returns
        -------
        object:
            An object used by the listeners.
        """
        return data

    def __dispatch_listeners(self, obj):
        """Dispatches the listeners waiting for the object.

        Parameters
        ----------
        obj: object
            The object to send to the listeners.
        """
        def notify():
            for listener in self.__listeners:
                if listener.accept(connection=self, message=obj):
                    listener.receive(connection=self, message=obj)
        threading.Thread(target=notify).start()

    def __listen(self, buffer_size=4096):
        """Listens to incoming data from the socket.

        Parameters
        ----------
        buffer_size: int
            The number of bytes for the socket to receive.
            Default value is `4096`.
        """

        trailing = b''
        while self.__is_connection_alive:
            data = self.__socket.recv(buffer_size)
            if data:

                # Handle messages which are only CRLFs.
                if data == b'\r\n':
                    self.__dispatch_listeners(self._process_data(trailing))
                    trailing = b''

                # Messages are separated by CR-LF. Last element is removed since it will be empty.
                split = data.split(b'\r\n')
                trailing += split.pop()

                # Ensure trailing information is added to the start of the next message.
                msg_count = len(split)
                if msg_count > 0:
                    split[0] = trailing + split[0]
                    trailing = b''

                for i in range(msg_count):
                    msg = split[i]
                    if msg:
                        self.__dispatch_listeners(self._process_data(msg))
            else:
                # Connection terminated by server: data was empty.
                self.__is_connection_alive = False


class MessageListener(object):

    """
    Listens for messages, and uses a filter to determine if the message should be accepted by the listener.
    If the message should be accepted, the implementation should then call MessageListener#receive.
    """

    def __init__(self, receive, message_filter=None):
        """
        Creates the listener.

        Parameters
        ----------
        receive: (Connection, object) -> None
            A method which takes a message as a parameter.
            This should only be called after message_filter returns True.
        message_filter: (Connection, object|None) -> bool
            A method which takes a connection and a message as parameters,
            and returns if the message should be accepted or not.
            If the `message_filter` is `None`, all messages will be accepted and passed to `receive`.
            Default value of `message_filter` is `None`.
        """
        self.__receive = receive
        self.__filter = message_filter

    def accept(self, connection, message):
        """
        Calls the `message_filter` parameter passed into the constructor.

        Parameters
        ----------
        connection: Connection
            The connection the message was sent over.
        message: object
            The message received.

        Returns
        -------
        bool: If the message should be accepted.
        """
        return self.__filter is None or self.__filter(connection, message)

    def receive(self, connection, message):
        """
        Calls the `receive` parameter passed into the constructor.

        Parameters
        ----------
        connection: Connection
            The connection the message was sent over.
        message: object
            The message being received.
        """
        self.__receive(connection, message)
